/*
Dual Stdio Code

This code allows serial port data to be accessed remotely
using the network.

Original author unknown. Modified by Craig A. Lindley
*/

#include "predef.h" 
#include <stdio.h>
#include <ctype.h>
#include <startnet.h>
#include <tcp.h>
#include <iointernal.h>
#include <serial.h>

#define RX_BUFFER_SIZE (2048)

int dual_fd;

static char receive_buffer[RX_BUFFER_SIZE];
static int serial_fd;
static int listen_fd;
static int telnet_fd;
static volatile int rx_put;
static volatile int rx_get;
static OS_CRIT dual_crit;
static bool bShare;

IoExpandStruct our_expand_funcs;


void reset_key_queue() {
	
	// Lock the critical section returning from function will clear this up as
	// the object destructor will be called.
	OSCriticalSectionObj crit(dual_crit);

	ClrDataAvail(dual_fd);
	SetWriteAvail(dual_fd);
	ClrHaveError(dual_fd);
	
	// Make queue empty
	rx_get = rx_put = 0;
}

int dual_read(int fd, char *buf, int nbytes) {

	int nbuf = 0;
	
	// Lock the critical section returning from function will clear this up as
	// the object destructor will be called.
	OSCriticalSectionObj crit(dual_crit);
	
	while ((nbuf < nbytes) && (rx_get != rx_put)) {
		buf[nbuf++] = receive_buffer[rx_get++];
		if (rx_get == RX_BUFFER_SIZE) {
			rx_get = 0;
		}
	}
	if (rx_get == rx_put) {
		ClrDataAvail(dual_fd);
	}
	return nbuf;
}

int dual_write(int fd, const char *buf, int nbytes) {
	
	if (telnet_fd) {
		int rv = write(telnet_fd, buf, nbytes);
		if(rv <= 0) {
			close(telnet_fd);
			telnet_fd = 0;
		}
	}
	if ((!telnet_fd) || (bShare)) {
		writeall(serial_fd, buf, nbytes);
	}
	return nbytes;
}


int dual_close(int fd) {
	
	// Do nothing we really don't let you close stdio
	return 0;
}

int ReadDualDataFrom(int fd) {
	
	OSCriticalSectionObj crit(dual_crit);
	
	int nread = 0;
	
	// Could be more efficent and read things in larger buffered sets by
	// doing the math around the buffer. This is just stdio so I'm going to
	// be really lazy and read one char at a time.
	while (dataavail(fd)) {
		char buf[1];
		
		int rv = read(fd, buf, 1);
		if (rv) {
			nread++;
			receive_buffer[rx_put++] = buf[0];
			if (rx_put == RX_BUFFER_SIZE) {
				rx_put = 0;
			}
			if (rx_get == rx_put) {
				//We have wrapped around so bump rx_get loosing the oldest char
				rx_get++;
				if (rx_get == RX_BUFFER_SIZE) {
					rx_get = 0;
				}
			}
		}
	}
	if (nread) {
		SetDataAvail(dual_fd);
	}
	return nread;
}

void DualStdioTask(void *p) {
	
	iprintf("Starting DualStdio Task\n");

	while (true) {
		fd_set fd_read;
		fd_set fd_error;
		FD_ZERO(&fd_read);
		FD_ZERO(&fd_error);
		
		FD_SET(listen_fd, &fd_read);
		if (telnet_fd) {
			FD_SET(telnet_fd, &fd_read);
			FD_SET(telnet_fd, &fd_error);
		}
		FD_SET(serial_fd, &fd_read);
		
		select(32, &fd_read, (fd_set *) NULL, &fd_error, 0);
		
		if (telnet_fd) {
			if (FD_ISSET(telnet_fd, &fd_error)) {
				//We got an error
				close(telnet_fd);
				telnet_fd = 0;
			}	else if (FD_ISSET(telnet_fd, &fd_read)) {
				if (ReadDualDataFrom(telnet_fd) == 0) {
					//Read zero bytes so we have been closed
					close(telnet_fd);
					telnet_fd = 0;
				}
			}
		}
		if (FD_ISSET(listen_fd, &fd_read)) {
			int new_fd = accept(listen_fd, NULL, NULL, 5);
			if (new_fd) {
				//For now we just bump off the old telnet fd
				if (telnet_fd) {
					close(telnet_fd);
					telnet_fd = 0;
				}
				telnet_fd = new_fd;
			}
		}
		if (FD_ISSET(serial_fd, &fd_read)) {
			if ((!telnet_fd) || (bShare)) {
				// No telnet or sharing read chars...
				ReadDualDataFrom(serial_fd);
			}	else	{
				// We have a valid telnet connection and we are set to not
				// share so we need to toss chars...
				// If we have more than 80 chars to read the select will
				// come by again.
				char buffer[80];
				read(serial_fd, buffer, 80);
			}
		}
	}
}

void CoreDualStdio(int serial_port,int baud_rate, int telnet_port, BYTE prio, BOOL share) {
		
	OSTimeDly(10);

	SerialClose(serial_port);
	serial_fd = SimpleOpenSerial(serial_port, baud_rate);
	listen_fd = listen(INADDR_ANY, telnet_port);
		
	OSCritInit(&dual_crit);
		
	our_expand_funcs.read = dual_read;
	our_expand_funcs.write = dual_write;
	our_expand_funcs.close = dual_close;
	
	dual_fd = GetExtraFD(NULL, &our_expand_funcs);
		
	bShare=share;
		
	ClrDataAvail(dual_fd);
	SetWriteAvail(dual_fd);
	ClrHaveError(dual_fd);
		
	ReplaceStdio(0, dual_fd);
	ReplaceStdio(1, dual_fd);
	ReplaceStdio(2, dual_fd);
		
	OSSimpleTaskCreate(DualStdioTask, prio);

	OSTimeDly(4);
}

// Set up so all stdio will go in/out of the telnet port if active and the serial port if not.
// setting baudrate to zero will cause it to use the config baud rate
// setting the telnet port to 0 will default to telnet (23)
// Only one or the other is active at any one time.
void StartDualStdio(int serial_port, int baud_rate, int telnet_port, BYTE prio) {
	
	CoreDualStdio(serial_port, baud_rate, telnet_port, prio, false);
}

